from __future__ import annotations

from typing import Any
from typing import Dict
from typing import List

import z3


class dNode:
    def translate(self, mappings: dict[str, str], types: dict[str, Any]):
        raise NotImplementedError

    def eval(self, x):
        raise NotImplementedError

    def collect_vars(self):
        raise NotImplementedError


class dLit(dNode):
    def __init__(self, val):
        self.val = val

    def translate(self, mappings, types):
        return self.val

    def eval(self, x):
        return self.val

    def __str__(self):
        return str(self.val)

    def collect_vars(self):
        return []


def s(x):
    """
    Sanitize ints
    :param x: to sanitize
    :return: sanitized
    """
    if isinstance(x, int) or isinstance(x, float):
        return dLit(x)

    return x


class dVar(dNode):

    # Name needed for multiple variables
    def __init__(self, names: list[str]):
        self.names = names

    def translate(self, mappings, types):
        resolved = mappings[self.names[0]]
        fullname = "_".join([resolved] + self.names[1:])
        cons = types[fullname]
        return cons(fullname)

    def eval(self, x):
        return x

    def __str__(self):
        return ".".join(self.names)

    def collect_vars(self):
        return [self.translate()]


class dAnd(dNode):
    def __init__(self, left, right):
        self.left = s(left)
        self.right = s(right)

    def translate(self, mappings, types):
        return z3.And(
            self.left.translate(
                mappings,
                types,
            ),
            self.right.translate(mappings, types),
        )

    def eval(self, x):
        return self.left.eval(x) and self.right.eval(x)

    def __str__(self):
        return f"{self.left} && {self.right}"

    def collect_vars(self):
        return self.left.collect_vars() + self.right.collect_vars()


class dOr(dNode):
    def __init__(self, left, right):
        self.left = s(left)
        self.right = s(right)

    def translate(self, mappings, types):
        return z3.Or(
            self.left.translate(
                mappings,
                types,
            ),
            self.right.translate(mappings, types),
        )

    def eval(self, x):
        return self.left.eval(x) or self.right.eval(x)

    def __str__(self):
        return f"{self.left} || {self.right}"

    def collect_vars(self):
        return self.left.collect_vars() + self.right.collect_vars()


class dLE(dNode):
    def __init__(self, left, right):
        self.left = s(left)
        self.right = s(right)

    def translate(self, mappings, types):
        return self.left.translate(mappings, types) <= self.right.translate(
            mappings,
            types,
        )

    def eval(self, x):
        return self.left.eval(x) <= self.right.eval(x)

    def __str__(self):
        return f"{self.left} <= {self.right}"

    def collect_vars(self):
        return self.left.collect_vars() + self.right.collect_vars()


class dLt(dNode):
    def __init__(self, left, right):
        self.left = s(left)
        self.right = s(right)

    def translate(self, mappings, types):
        return self.left.translate(mappings, types) < self.right.translate(
            mappings,
            types,
        )

    def eval(self, x):
        return self.left.eval(x) < self.right.eval(x)

    def __str__(self):
        return f"{self.left} < {self.right}"

    def collect_vars(self):
        return self.left.collect_vars() + self.right.collect_vars()


class dGE(dNode):
    def __init__(self, left, right):
        self.left = s(left)
        self.right = s(right)

    def translate(self, mappings, types):
        return self.left.translate(mappings, types) >= self.right.translate(
            mappings,
            types,
        )

    def eval(self, x):
        return self.left.eval(x) >= self.right.eval(x)

    def __str__(self):
        return f"{self.left} >= {self.right}"

    def collect_vars(self):
        return self.left.collect_vars() + self.right.collect_vars()


class dGt(dNode):
    def __init__(self, left, right):
        self.left = s(left)
        self.right = s(right)

    def translate(self, mappings, types):
        return self.left.translate(mappings, types) > self.right.translate(
            mappings,
            types,
        )

    def eval(self, x):
        return self.left.eval(x) > self.right.eval(x)

    def __str__(self):
        return f"{self.left} > {self.right}"

    def collect_vars(self):
        return self.left.collect_vars() + self.right.collect_vars()


class dEQ(dNode):
    def __init__(self, left, right):
        self.left = s(left)
        self.right = s(right)

    def translate(self, mappings, types):
        return self.left.translate(mappings, types) == self.right.translate(
            mappings,
            types,
        )

    def eval(self, x):
        return self.left.eval(x) == self.right.eval(x)

    def __str__(self):
        return f"{self.left} == {self.right}"

    def collect_vars(self):
        return self.left.collect_vars() + self.right.collect_vars()


class dMod(dNode):
    def __init__(self, left, right):
        self.left = s(left)
        self.right = s(right)

    def translate(self, mappings, types):
        return self.left.translate(mappings, types) % self.right.translate(
            mappings,
            types,
        )

    def eval(self, x):
        return self.left.eval(x) % self.right.eval(x)

    def __str__(self):
        return f"{self.left} % {self.right}"

    def collect_vars(self):
        return self.left.collect_vars() + self.right.collect_vars()


class dPlus(dNode):
    def __init__(self, left, right):
        self.left = s(left)
        self.right = s(right)

    def translate(self, mappings, types):
        return self.left.translate(mappings, types) + self.right.translate(
            mappings,
            types,
        )

    def eval(self, x):
        return self.left.eval(x) + self.right.eval(x)

    def __str__(self):
        return f"({self.left} + {self.right})"

    def collect_vars(self):
        return self.left.collect_vars() + self.right.collect_vars()


class dNEQ(dNode):
    def __init__(self, left, right):
        self.left = s(left)
        self.right = s(right)

    def translate(self, mappings, types):
        return self.left.translate(mappings, types) != self.right.translate(
            mappings,
            types,
        )

    def eval(self, x):
        return self.left.eval(x) != self.right.eval(x)

    def __str__(self):
        return f"{self.left} != {self.right}"

    def collect_vars(self):
        return self.left.collect_vars() + self.right.collect_vars()


class dNot(dNode):
    def __init__(self, cond):
        self.cond = s(cond)

    def translate(self, mappings, types):
        return z3.Not(self.cond.translate(mappings, types))

    def eval(self, x):
        return not self.cond.eval(x)

    def __str__(self):
        return f"!({self.cond})"

    def collect_vars(self):
        return self.cond.collect_vars()


class fNavigate(dNode):
    def __init__(self, obj, expr):
        self.obj = obj
        self.expr = expr

    def translate_mappings(self, mappings: dict[str, str]):
        ret = {}
        obj = self.obj
        for k, v in mappings.items():
            if k.startswith(obj + "."):
                newk = k[len(obj) + 1 :]
            else:
                newk = "__parent__." + k
            ret[newk] = v
        return ret

    def translate(self, mappings: dict[str, str], types):
        mappings = self.translate_mappings(mappings)

        return self.expr.translate(mappings, types)

    def eval(self, x):
        raise NotImplementedError()

    def __str__(self):
        return self.obj + "." + str(self.expr)

    def collect_vars(self):
        raise NotImplementedError()


class dAllPairs(dNode):
    def __init__(self, cont: str, x: str, y: str, block: dNode):
        self.cont = cont
        self.x = x
        self.y = y
        self.block = block

    def _gen_pairs(self, l: list):
        for i1, e1 in enumerate(l):
            for i2, e2 in enumerate(l):
                if i1 != i2:
                    yield e1, e2

    def translate(self, mappings: dict[str, str], types: dict[str, Any]):
        real_cont = mappings[self.cont]
        subs = []
        for k in types.keys():
            if not k.startswith(real_cont):
                continue
            trim_k = k[len(real_cont) :]
            if "." in trim_k:
                continue
            subs.append(k)

        ret = True

        for x, y in self._gen_pairs(subs):
            maps = mappings.copy()
            maps[self.x] = x
            maps[self.y] = y
            transed = self.block.translate(maps, types)
            ret = z3.And(ret, transed)

        return ret



from __future__ import annotations

from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Type
from typing import TypeVar

from geneticengine.core.grammar import Grammar
from geneticengine.core.random.sources import Source
from geneticengine.metahandlers.base import MetaHandlerGenerator

T = TypeVar("T")


class VarRange(MetaHandlerGenerator):
    """
    VarRange([a, b, c]) represents the alternative between a, b, and c.
    The list of options can be dynamically altered before the grammar extraction
    with something like Var.__init__.__annotations__["name"] = Annotated[str, VarRange([d, e, f])].
    The option list must not be empty.
    """

    def __init__(self, options: list[T]):
        if not options:
            raise Exception(
                f"The VarRange metahandler requires a non-empty set of options. Options found: {options}",
            )
        self.options = options

    def generate(
        self,
        r: Source,
        g: Grammar,
        rec,
        new_symbol,
        depth: int,
        base_type,
        context: dict[str, str],
    ):
        rec(r.choice(self.options, str(base_type)))

    def __repr__(self):
        return str(self.options)

    def __class_getitem__(self, args):
        return VarRange(*args)



from __future__ import annotations

from dataclasses import dataclass
from typing import Any

from geneticengine.core.decorators import get_gengy
from geneticengine.core.grammar import Grammar
from geneticengine.core.random.sources import Source
from geneticengine.core.representations.api import Representation
from geneticengine.core.representations.tree.initialization_methods import Initialization_Method
from geneticengine.core.representations.tree.treebased import PI_Grow
from geneticengine.core.representations.tree.treebased import random_node
from geneticengine.core.tree import TreeNode
from geneticengine.core.utils import get_arguments
from geneticengine.core.utils import get_generic_parameter
from geneticengine.core.utils import is_generic
from geneticengine.core.utils import is_generic_list
from geneticengine.core.utils import strip_annotations
from geneticengine.exceptions import GeneticEngineError
from geneticengine.metahandlers.base import is_metahandler

MAX_RAND_INT = 100000
MAX_RAND_LIST_SIZE = 10


LEFTOVER_KEY = "$leftovers"


@dataclass
class Genotype:
    dna: dict[str, list[int]]

    def register_production(self, prod_index: int, starting_symbol):
        if str(starting_symbol) in self.dna.keys():
            self.dna[str(starting_symbol)].append(prod_index)
        else:
            self.dna[str(starting_symbol)] = [prod_index]


def filter_choices(possible_choices: list[type], g: Grammar, depth, starting_symbol):
    valid_productions = [
        vp for vp in possible_choices if g.get_distance_to_terminal(vp) <= depth
    ]
    if not valid_productions:
        raise GeneticEngineError(
            "No productions for non-terminal node with type: {} in depth {} (minimum required: {}).".format(
                starting_symbol,
                depth,
                str(
                    [(vp, g.distanceToTerminal[vp]) for vp in possible_choices],
                ),
            ),
        )
    return valid_productions


def assert_depth_error(max_depth, g, starting_symbol):
    if max_depth < g.get_distance_to_terminal(starting_symbol):
        raise GeneticEngineError(
            "There will be no depth sufficient for symbol {} in this grammar (provided: {}, required: {}).".format(
                starting_symbol,
                max_depth,
                g.get_distance_to_terminal(
                    starting_symbol,
                ),
            ),
        )


def random_individual(
    r: Source,
    g: Grammar,
    starting_symbol: Any,
    current_genotype: Genotype | None = None,
    max_depth: int = 5,
) -> Genotype: 
    if current_genotype == None:
        nodes = [str(node) for node in g.all_nodes]
        for node in g.all_nodes:
            arguments = get_arguments(node)
            for _, arg in arguments:
                if is_generic(arg):
                    nodes.append(str(arg))
                base_type = str(strip_annotations(arg))
                if base_type not in nodes:
                    nodes.append(base_type)

        dna: dict[str, list[int]] = dict()
        for nodestr in nodes:
            dna[nodestr] = list()
        dna[LEFTOVER_KEY] = [
            r.randint(0, MAX_RAND_INT) for _ in range(1000)
        ]  # Necessary to source from when a production rule runs out of genes.
        current_genotype = Genotype(dna)
    assert type(current_genotype) == Genotype

    if starting_symbol in [int, float, str, bool]:
        val = r.randint(0, MAX_RAND_INT)
        current_genotype.register_production(val, starting_symbol)
    elif starting_symbol in g.alternatives:
        assert_depth_error(max_depth, g, starting_symbol)
        productions = g.alternatives[starting_symbol]
        valid_productions = filter_choices(
            productions,
            g,
            max_depth - int(g.expansion_depthing),
            starting_symbol,
        )
        if any(["weight" in get_gengy(p) for p in valid_productions]):
            weights = [get_gengy(p).get("weight", 1.0) for p in valid_productions]
            prod = r.choice_weighted(
                valid_productions,
                weights,
                str(starting_symbol),
            )
        else:
            prod = r.choice(valid_productions)
        prod_index = valid_productions.index(prod)
        current_genotype.register_production(prod_index, starting_symbol)
        current_genotype = random_individual(
            r,
            g,
            prod,
            current_genotype,
            max_depth - int(g.expansion_depthing),
        )
    elif is_generic_list(starting_symbol):
        new_type = get_generic_parameter(starting_symbol)
        list_length = r.randint(1, MAX_RAND_LIST_SIZE)
        current_genotype.register_production(list_length, int)
        for _ in range(list_length):
            current_genotype = random_individual(
                r,
                g,
                new_type,
                current_genotype,
                max_depth,
            )
    elif is_metahandler(
        starting_symbol,
    ):  # No need to go down one in depth as this is the argument of a treenode and therefore already has an adjusted depth.
        if is_generic_list(get_generic_parameter(starting_symbol)):
            new_type = get_generic_parameter(starting_symbol)
            current_genotype = random_individual(
                r,
                g,
                new_type,
                current_genotype,
                max_depth,
            )
        else:
            new_type = strip_annotations(starting_symbol)
            current_genotype = random_individual(
                r,
                g,
                new_type,
                current_genotype,
                max_depth,
            )
    else:
        assert_depth_error(max_depth, g, starting_symbol)
        if starting_symbol not in g.all_nodes:
            raise GeneticEngineError(
                f"Symbol {starting_symbol} not in grammar rules.",
            )

        args = get_arguments(starting_symbol)
        for _, argt in args:
            current_genotype = random_individual(
                r,
                g,
                argt,
                current_genotype,
                max_depth - 1,
            )

    return current_genotype



def create_individual(
    r: Source,
    g: Grammar,
    starting_symbol: Any = None,
    current_genotype: Genotype | None = None,
    max_depth: int = 5,
) -> Genotype:
    if not starting_symbol:
        starting_symbol = g.starting_symbol

    return random_individual(r, g, starting_symbol, current_genotype, max_depth)


def mutate(r: Source, g: Grammar, ind: Genotype, max_depth: int) -> Genotype:
    rkey = r.choice(
        list(
            key
            for key in ind.dna.keys()
            if (len(ind.dna[key]) > 0) and (key != LEFTOVER_KEY)
        ),
    )
    dna = ind.dna
    clone = [i for i in dna[rkey]]
    rindex = r.randint(0, len(dna[rkey]) - 1)
    clone[rindex] = r.randint(0, MAX_RAND_INT)
    dna[rkey] = clone
    return Genotype(dna)


def crossover(
    r: Source,
    g: Grammar,
    p1: Genotype,
    p2: Genotype,
    max_depth: int,
) -> tuple[Genotype, Genotype]:
    keys = p1.dna.keys()  # Leave leftovers in as it doesn't matter
    mask = [(k, r.random_bool()) for k in keys]
    c1 = dict()
    c2 = dict()
    for k, b in mask:
        if b:
            c1[k] = p1.dna[k]
            c2[k] = p2.dna[k]
        else:
            c1[k] = p2.dna[k]
            c2[k] = p1.dna[k]
    return (Genotype(c1), Genotype(c2))


class DynamicStructuredListWrapper(Source):
    ind: Genotype
    indexes: dict[str, int]

    def __init__(self, ind: Genotype):
        self.ind = ind
        indexes = dict()
        for k in ind.dna.keys():
            indexes[k] = 0
        self.indexes = indexes

    def register_index(self, prod):
        if prod in self.indexes.keys():
            self.indexes[prod] += 1
        else:
            self.ind.dna[prod] = list()
            self.indexes[prod] = 1

    def randint(self, min: int, max: int, prod: str = "") -> int:
        self.register_index(prod)
        if self.indexes[prod] >= len(
            self.ind.dna[prod],
        ):  # We don't have a wrapper function, but we add elements to each list when there are no genes left. These are sourced from the "left_overs" in the dna.
            self.indexes[LEFTOVER_KEY] = (self.indexes[LEFTOVER_KEY] + 1) % len(
                self.ind.dna[LEFTOVER_KEY],
            )
            self.ind.register_production(
                self.ind.dna[LEFTOVER_KEY][self.indexes[LEFTOVER_KEY]],
                prod,
            )
        v = self.ind.dna[prod][self.indexes[prod] - 1]
        return v % (max - min + 1) + min

    def random_float(self, min: float, max: float, prod: str = "") -> float:
        k = self.randint(1, 100000000, prod)
        return 1 * (max - min) / k + min


def create_tree(g: Grammar, ind: Genotype, depth: int, method) -> TreeNode:
    rand: Source = DynamicStructuredListWrapper(ind)
    return random_node(rand, g, depth, g.starting_symbol, method=method)


class DynamicStructuredGrammaticalEvolutionRepresentation(Representation[Genotype]):
    """This version uses a list of lists of integers to represent individuals, based on non-terminal
    symbols. It delays computing the expansions that have enough depth to runtime."""

    def __init__(self, depth = None, method: Initialization_Method = PI_Grow()) -> None:
        self.depth = depth
        self.method = method
    
    def create_individual(self, r: Source, g: Grammar, depth: int) -> Genotype:
        self.depth = depth
        return create_individual(r, g, max_depth=depth)

    def mutate_individual(
        self,
        r: Source,
        g: Grammar,
        ind: Genotype,
        depth: int,
        ty: type,
        specific_type: type | None = None,
        depth_aware_mut: bool = False,
    ) -> Genotype:
        return mutate(r, g, ind, depth)

    def crossover_individuals(
        self,
        r: Source,
        g: Grammar,
        i1: Genotype,
        i2: Genotype,
        depth: int,
        specific_type: type | None = None,
        depth_aware_co: bool = False,
    ) -> tuple[Genotype, Genotype]:
        return crossover(r, g, i1, i2, depth)

    def genotype_to_phenotype(self, g: Grammar, genotype: Genotype) -> TreeNode:
        return create_tree(g, genotype, self.depth, self.method.tree_init_method)


dsge_representation = DynamicStructuredGrammaticalEvolutionRepresentation()



from __future__ import annotations

import weakref
from collections import defaultdict
from typing import Any
from typing import Dict
from typing import List
from typing import Set
from typing import Tuple

from geneticengine.core.decorators import is_builtin
from geneticengine.core.grammar import Grammar
from geneticengine.core.tree import TreeNode
from geneticengine.core.utils import get_arguments
from geneticengine.core.utils import is_abstract
from geneticengine.core.utils import is_terminal


class GengyList(list):
    def __init__(self, typ, vals):
        super().__init__(vals)
        self.typ = typ
        self.gengy_init_values = vals

    def new_like(self, *newargs):
        return GengyList(self.typ, newargs)


def relabel_nodes(
    i: TreeNode,
    g: Grammar,
    is_list: bool = False,
) -> tuple[int, int, dict[type, list[Any]], int]:
    """
    Recomputes all the nodes, depth and distance_to_term in the tree.\n
    Returns the number of nodes, distance to terminal (depth), typed this way, and the weighted number of nodes (counting depth points instead of nodes).
    """
    non_terminals = g.non_terminals
    children: list[Any]
    if getattr(i, "gengy_labeled", False):
        return (
            i.gengy_nodes,
            i.gengy_distance_to_term,
            i.gengy_types_this_way,
            i.gengy_weighted_nodes,
        )
    number_of_nodes = 1
    distance_to_term = 1
    weighted_number_of_nodes = 0
    if is_list:
        number_of_nodes = 0
        distance_to_term = 0
    types_this_way = defaultdict(lambda: [])
    types_this_way[type(i)] = [i]
    if is_terminal(type(i), non_terminals) and (not isinstance(i, list)):
        if not is_builtin(type(i)):
            i.gengy_labeled = True
            i.gengy_distance_to_term = int(g.expansion_depthing)
            i.gengy_nodes = int(g.expansion_depthing)
            i.gengy_weighted_nodes = int(g.expansion_depthing)
            i.gengy_types_this_way = {type(i): [i]}
        return (
            int(g.expansion_depthing),
            int(g.expansion_depthing),
            {type(i): [i]},
            int(g.expansion_depthing),
        )
    else:
        if isinstance(i, list):
            children = [(type(obj), obj) for obj in i]
        else:
            if not hasattr(i, "gengy_init_values"):
                breakpoint()
            children = [
                (typ[1], i.gengy_init_values[idx])
                for idx, typ in enumerate(get_arguments(i))
            ]
        assert children
        for t, c in children:
            nodes, dist, thisway, weighted_nodes = relabel_nodes(
                c,
                g,
                isinstance(c, list),
            )
            abs_adjust = (
                0
                if not is_abstract(t) or not g.expansion_depthing
                else g.abstract_dist_to_t[t][type(c)]
            )
            if isinstance(c, list) and g.expansion_depthing:
                abs_adjust = 1
            list_adjust = 0 if isinstance(c, list) else 1
            number_of_nodes += abs_adjust + nodes
            weighted_number_of_nodes += weighted_nodes
            distance_to_term = max(distance_to_term, dist + abs_adjust + list_adjust)
            for (k, v) in thisway.items():
                types_this_way[k].extend(v)

    if not is_list:
        weighted_number_of_nodes += distance_to_term

    i.gengy_labeled = True
    i.gengy_distance_to_term = distance_to_term
    i.gengy_nodes = number_of_nodes
    i.gengy_weighted_nodes = weighted_number_of_nodes
    i.gengy_types_this_way = types_this_way
    return number_of_nodes, distance_to_term, types_this_way, weighted_number_of_nodes


def relabel_nodes_of_trees(i: TreeNode, g: Grammar) -> TreeNode:
    """Recomputes all the nodes, depth and distance_to_term in the tree"""

    relabel_nodes(i, g)
    return i



from __future__ import annotations

from abc import ABC
from dataclasses import dataclass
from typing import Annotated

from geneticengine.metahandlers.ints import IntRange
from geneticengine.metahandlers.vars import VarRange


class Number(ABC):
    """
    Standard Number object.
    """

    def evaluate(self, **kwargs):
        return 0.0


@dataclass
class Plus(Number):
    """
    Standard Plus object.

    Args:
        left  (Number)
        right (Number)

    Returns when evaluated:
        left + right
    """

    left: Number
    right: Number

    def evaluate(self, **kwargs):
        return self.left.evaluate(**kwargs) + self.right.evaluate(**kwargs)

    def __str__(self) -> str:
        return f"({self.left} + {self.right})"


@dataclass
class Minus(Number):
    """
    Standard Minus object.

    Args:
        left  (Number)
        right (Number)

    Returns when evaluated:
        left - right
    """

    left: Number
    right: Number

    def evaluate(self, **kwargs):
        return self.left.evaluate(**kwargs) - self.right.evaluate(**kwargs)

    def __str__(self) -> str:
        return f"({self.left} - {self.right})"


@dataclass
class Mul(Number):
    """
    Standard Multiplication object.

    Args:
        left  (Number)
        right (Number)

    Returns when evaluated:
        left * right
    """

    left: Number
    right: Number

    def evaluate(self, **kwargs):
        return self.left.evaluate(**kwargs) * self.right.evaluate(**kwargs)

    def __str__(self) -> str:
        return f"({self.left} * {self.right})"


@dataclass
class Literal(Number):
    """
    Standard Literal object.

    Args:
        val  (Number)

    Returns when evaluated:
        val
    """

    val: Annotated[int, IntRange(0, 9)]

    def evaluate(self, **kwargs):
        return self.val

    def __str__(self) -> str:
        return str(self.val)


@dataclass
class Var(Number):
    """
    Standard Variable object. Used to introduce variables.

    Args:
        name  (str)

    Returns when evaluated:
        name
    """

    name: Annotated[str, VarRange(["x", "y", "z"])]

    def evaluate(self, **kwargs):
        return kwargs[self.name]

    def __str__(self) -> str:
        return self.name



from __future__ import annotations

from abc import ABC
from abc import abstractmethod
from operator import attrgetter
from typing import Any
from typing import Callable

from geneticengine.algorithms.gp.individual import Individual
from geneticengine.core.grammar import Grammar
from geneticengine.core.problems import FitnessType
from geneticengine.core.problems import MultiObjectiveProblem
from geneticengine.core.problems import Problem
from geneticengine.core.problems import SingleObjectiveProblem
from geneticengine.core.random.sources import RandomSource
from geneticengine.core.representations.api import Representation


class Heuristics(ABC):
    """
    An Abstract class that gp.py, hill_climbing.py and random_search extends to

    Args:
        grammar (Grammar): The grammar used to guide the search.
        representation (Representation): The individual representation used by the GP program. The default is treebased_representation.
        problem (Problem): The problem we are solving. Either a SingleObjectiveProblem or a MultiObjectiveProblem.
        randomSource (Callable[[int], RandomSource]): The random source function used by the program. Should take in an integer, representing the seed, and return a RandomSource.

    """

    grammar: Grammar
    representation: Representation
    problem: Problem
    random: RandomSource

    def __init__(
        self,
        grammar: Grammar,
        representation: Representation,
        problem: Problem,
        randomSource: Callable[[int], RandomSource],
        seed: int,
    ):
        self.problem: Problem = problem
        self.grammar = grammar
        self.representation = representation
        self.random = randomSource(seed)

    @abstractmethod
    def evolve(self, verbose):
        ...

    def get_best_individual(
        self,
        p: Problem,
        individuals: list[Individual],
    ) -> Individual:
        """
        The get_best_individual is a method that that returns the best individual of a population

        Args:
            p (Problem): the problem we are trying to solve
            individuals (list[Individual]): the list of individuals where we're going to search for the best one

        Returns:
            returns an Individual
        """
        best_individual: Individual
        fitnesses = [self.evaluate(x) for x in individuals]

        if isinstance(p, SingleObjectiveProblem):
            assert all(isinstance(x, float) for x in fitnesses)
            if p.minimize:
                best_individual = min(individuals, key=attrgetter("fitness"))
            else:
                best_individual = max(individuals, key=attrgetter("fitness"))

        elif isinstance(p, MultiObjectiveProblem):
            assert all(isinstance(x, list) for x in fitnesses)

            def single_criteria(i: Individual) -> float:
                assert isinstance(p.minimize, list)
                assert isinstance(i.fitness, list)
                return sum((m and -f or f) for (f, m) in zip(i.fitness, p.minimize))

            if p.best_individual_criteria_function is None:
                best_individual = max(individuals, key=single_criteria)
            else:
                best_individual = max(
                    individuals,
                    key=p.best_individual_criteria_function,
                )

        return best_individual

    # this only works with SingleObjectiveProblem
    def keyfitness(self):
        if self.problem.minimize:
            return lambda x: self.evaluate(x)
        else:
            return lambda x: -self.evaluate(x)

    def evaluate(self, individual: Individual) -> FitnessType:
        """
        The evaluate is a methoad that is used to evaluate individual fitness

        Args:
            individual (Individual): Individual that we are evaluating the fitness

        Returns:
            FitnessType: The FitnessType of the individual, either a float or a list[float]

        """
        if individual.fitness is None:
            phenotype = self.representation.genotype_to_phenotype(
                self.grammar,
                individual.genotype,
            )
            individual.fitness = self.problem.evaluate(phenotype)
        return individual.fitness

    def create_individual(self, depth: int) -> Individual:
        """
        The create_individual is a methoad that is used to create a new individual

        Args:
            depth: number of

        Returns:
            Individual

        """
        genotype = self.representation.create_individual(
            r=self.random,
            g=self.grammar,
            depth=depth,
        )
        return Individual(
            genotype=genotype,
            fitness=None,
        )



from __future__ import annotations

from abc import abstractmethod
from dataclasses import dataclass
from typing import Any
from typing import Callable
from typing import Generic
from typing import Tuple
from typing import Type
from typing import TypeVar

from geneticengine.core.grammar import Grammar
from geneticengine.core.random.sources import Source
from geneticengine.core.representations.tree.initialization_methods import Initialization_Method
from geneticengine.core.tree import TreeNode

g = TypeVar("g")


class Representation(Generic[g]):
    depth: int
    method: Initialization_Method
    
    @abstractmethod
    def create_individual(self, r: Source, g: Grammar, depth: int) -> g:
        ...

    @abstractmethod
    def mutate_individual(
        self,
        r: Source,
        g: Grammar,
        ind: g,
        depth: int,
        ty: type,
        specific_type: type | None = None,
        depth_aware_mut: bool = False,
    ) -> g:
        ...

    @abstractmethod
    def crossover_individuals(
        self,
        r: Source,
        g: Grammar,
        i1: g,
        i2: g,
        int,
        specific_type: type | None = None,
        depth_aware_co: bool = False,
    ) -> tuple[g, g]:
        ...

    @abstractmethod
    def genotype_to_phenotype(self, g: Grammar, genotype: g) -> TreeNode:
        ...



from __future__ import annotations

from dataclasses import dataclass
from math import isinf
from typing import Annotated
from typing import List
from typing import Type

import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator
from sklearn.base import TransformerMixin

from geneticengine.algorithms.gp.gp import GP
from geneticengine.algorithms.hill_climbing import HC
from geneticengine.core.grammar import extract_grammar
from geneticengine.core.problems import Problem
from geneticengine.core.problems import SingleObjectiveProblem
from geneticengine.core.random.sources import RandomSource
from geneticengine.core.representations.api import Representation
from geneticengine.core.representations.tree.treebased import treebased_representation
from geneticengine.grammars.basic_math import SafeDiv
from geneticengine.grammars.basic_math import SafeLog
from geneticengine.grammars.basic_math import SafeSqrt
from geneticengine.grammars.literals import exp_literals
from geneticengine.grammars.literals import ExpLiteral
from geneticengine.grammars.sgp import Literal
from geneticengine.grammars.sgp import Mul
from geneticengine.grammars.sgp import Number
from geneticengine.grammars.sgp import Plus
from geneticengine.grammars.sgp import Var
from geneticengine.metahandlers.vars import VarRange
from geneticengine.metrics import f1_score
from geneticengine.off_the_shelf.sympy_compatible import fix_all


class GeneticProgrammingClassifier(BaseEstimator, TransformerMixin):
    """
    Genetic Programming Classifier. Main attributes: fit and predict
    Defaults as given in A Field Guide to GP, p.17, by Poli and Mcphee:

    Args:
        nodes (List[Number]): The list of nodes to be used in the grammar. You can design your own, or use the ones in geneticengine.grammars.[sgp,literals,basic_math]. The default uses [ Plus, Mul, ExpLiteral, Var, SafeDiv, SafeLog, SafeSqrt ] + exp_literals.
        representation (Representation): The individual representation used by the GP program. The default is treebased_representation. Currently Genetic Engine also supports Grammatical Evolution: geneticengine.core.representations.grammatical_evolution.ge_representation. You can also deisgn your own.
        seed (int): The seed for the RandomSource (default = 123).
        population_size (int): The population size (default = 200). Apart from the first generation, each generation the population is made up of the elites, novelties, and transformed individuals from the previous generation. Note that population_size > (n_elites + n_novelties + 1) must hold.
        n_elites (int): Number of elites, i.e. the number of best individuals that are preserved every generation (default = 5).
        n_novelties (int): Number of novelties, i.e. the number of newly generated individuals added to the population each generation. (default = 10).
        number_of_generations (int): Number of generations (default = 100).
        max_depth (int): The maximum depth a tree can have (default = 15).
        favor_less_complex_trees (bool): If set to True, this gives a tiny penalty to more complex (with more nodes) trees to favor simpler trees (default = False).
        hill_climbing (bool): Allows the user to change the standard mutation operations to the hill-climbing mutation operation, in which an individual is mutated to 5 different new individuals, after which the best is chosen to survive (default = False).

        probability_mutation (float): probability that an individual is mutated (default = 0.01).
        probability_crossover (float): probability that an individual is chosen for cross-over (default = 0.9).

    """

    def __init__(
        self,
        nodes: list[type[Number]] | None = None,
        representation: Representation = treebased_representation,
        population_size: int = 200,
        n_elites: int = 5,  # Shouldn't this be a percentage of population size?
        n_novelties: int = 10,
        number_of_generations: int = 100,
        max_depth: int = 15,
        favor_less_complex_trees: bool = True,
        hill_climbing: bool = False,
        seed: int = 123,
        # -----
        # As given in A Field Guide to GP, p.17, by Poli and Mcphee
        probability_mutation: float = 0.01,
        probability_crossover: float = 0.9,
        # -----
    ):
        if nodes is None:
            nodes = [
                Plus,
                Mul,
                ExpLiteral,
                Var,
                SafeDiv,
                SafeLog,
                SafeSqrt,
                *exp_literals,
            ]
        assert population_size > (n_elites + n_novelties + 1)
        assert Var in nodes

        self.nodes = nodes
        self.representation = representation
        self.evolved_ind = None
        self.nodes = nodes
        self.random = RandomSource(seed)
        self.seed = seed
        self.population_size = population_size
        self.max_depth = max_depth
        self.favor_less_complex_trees = favor_less_complex_trees
        self.n_elites = n_elites
        self.n_novelties = n_novelties
        self.hill_climbing = hill_climbing
        self.number_of_generations = number_of_generations
        self.probability_mutation = probability_mutation
        self.probability_crossover = probability_crossover

    def fit(self, X, y):
        """
        Fits the classifier with data X and target y.
        """
        if type(y) == pd.Series:
            target = y.values
        else:
            target = y

        if type(X) == pd.DataFrame:
            feature_names = list(X.columns.values)
            data = X.values
        else:
            feature_names = [f"x{i}" for i in range(X.shape[1])]
            data = X
        feature_indices = {}
        for i, n in enumerate(feature_names):
            feature_indices[n] = i

        Var.__init__.__annotations__["name"] = Annotated[str, VarRange(feature_names)]
        Var.feature_indices = feature_indices

        self.grammar = extract_grammar(self.nodes, Number)
        self.feature_names = feature_names
        self.feature_indices = feature_indices

        def fitness_function(n: Number):
            variables = {}
            for x in feature_names:
                i = feature_indices[x]
                variables[x] = data[:, i]
            y_pred = n.evaluate(**variables)

            if type(y_pred) in [np.float64, int, float]:
                """If n does not use variables, the output will be scalar."""
                y_pred = np.full(len(target), y_pred)
            if y_pred.shape != (len(target),):
                return -100000000
            fitness = f1_score(y_pred, target)
            if isinf(fitness):
                fitness = -100000000
            return fitness

        model = GP(
            grammar=self.grammar,
            problem=SingleObjectiveProblem(
                minimize=False,
                fitness_function=fitness_function,
                target_fitness=None,
            ),
            representation=self.representation,
            population_size=self.population_size,
            n_elites=self.n_elites,
            n_novelties=self.n_novelties,
            number_of_generations=self.number_of_generations,
            max_depth=self.max_depth,
            favor_less_complex_trees=self.favor_less_complex_trees,
            probability_mutation=self.probability_mutation,
            probability_crossover=self.probability_crossover,
            hill_climbing=self.hill_climbing,
            seed=self.seed,
        )

        best_ind, fitness, phenotype = model.evolve(verbose=1)
        self.evolved_phenotype = phenotype
        self.sympy_compatible_phenotype = fix_all(str(phenotype))

    def predict(self, X):
        """
        Predict the target values of X. The model must have been fitted
        """
        assert self.evolved_phenotype != None
        if (type(X) == pd.DataFrame) or (type(X) == pd.Series):
            data = X.values
        else:
            data = X
        if len(data.shape) == 1:
            data = np.array([data])
        assert data.shape[1] == len(self.feature_names)

        variables = {}
        for x in self.feature_names:
            i = self.feature_indices[x]
            variables[x] = data[:, i]
        y_pred = self.evolved_phenotype.evaluate(**variables)

        return y_pred


class HillClimbingClassifier(BaseEstimator, TransformerMixin):
    """
    Hill Climbing Classifier. Main attributes: fit and predict

    Args:
        nodes (List[Number]): The list of nodes to be used in the grammar. You can design your own, or use the ones in geneticengine.grammars.[sgp,literals,basic_math]. The default uses [ Plus, Mul, ExpLiteral, Var, SafeDiv, SafeLog, SafeSqrt ] + exp_literals.
        representation (Representation): The individual representation used by the GP program. The default is treebased_representation. Currently Genetic Engine also supports Grammatical Evolution: geneticengine.core.representations.grammatical_evolution.ge_representation. You can also deisgn your own.
        seed (int): The seed for the RandomSource (default = 123).
        population_size (int): The population size (default = 200). Apart from the first generation, each generation the population is made up of the elites, novelties, and transformed individuals from the previous generation. Note that population_size > (n_elites + n_novelties + 1) must hold.
        number_of_generations (int): Number of generations (default = 100).
        max_depth (int): The maximum depth a tree can have (default = 15).
    """

    def __init__(
        self,
        nodes=[
            Plus,
            Mul,
            ExpLiteral,
            Var,
            SafeDiv,
            SafeLog,
            SafeSqrt,
            *exp_literals,
        ],  # "type: ignore"
        representation: Representation = treebased_representation,
        population_size: int = 200,
        number_of_generations: int = 100,
        max_depth: int = 15,
        seed: int = 123,
    ):
        assert Var in nodes

        self.nodes = nodes
        self.representation = representation
        self.evolved_ind = None
        self.nodes = nodes
        self.random = RandomSource(seed)
        self.seed = seed
        self.population_size = population_size
        self.max_depth = max_depth
        self.number_of_generations = number_of_generations

    def fit(self, X, y):
        """
        Fits the classifier with data X and target y.
        """
        if type(y) == pd.Series:
            target = y.values
        else:
            target = y

        if type(X) == pd.DataFrame:
            feature_names = list(X.columns.values)
            data = X.values
        else:
            feature_names = [f"x{i}" for i in range(X.shape[1])]
            data = X
        feature_indices = {}
        for i, n in enumerate(feature_names):
            feature_indices[n] = i

        Var.__init__.__annotations__["name"] = Annotated[str, VarRange(feature_names)]
        Var.feature_indices = feature_indices

        self.grammar = extract_grammar(self.nodes, Number)
        self.feature_names = feature_names
        self.feature_indices = feature_indices

        def fitness_function(n: Number):
            variables = {}
            for x in feature_names:
                i = feature_indices[x]
                variables[x] = data[:, i]
            y_pred = n.evaluate(**variables)

            if type(y_pred) in [np.float64, int, float]:
                """If n does not use variables, the output will be scalar."""
                y_pred = np.full(len(target), y_pred)
            if y_pred.shape != (len(target),):
                return -100000000
            fitness = f1_score(y_pred, target)
            if isinf(fitness):
                fitness = -100000000
            return fitness

        model = HC(
            g=self.grammar,
            problem=SingleObjectiveProblem(
                minimize=False,
                fitness_function=fitness_function,
                target_fitness=None,
            ),
            representation=self.representation,
            population_size=self.population_size,
            number_of_generations=self.number_of_generations,
            max_depth=self.max_depth,
            seed=self.seed,
        )

        best_ind, fitness, phenotype = model.evolve(verbose=1)
        self.evolved_phenotype = phenotype
        self.sympy_compatible_phenotype = fix_all(str(phenotype))

    def predict(self, X):
        """
        Predict the target values of X. The model must have been fitted
        """
        assert self.evolved_phenotype != None
        if (type(X) == pd.DataFrame) or (type(X) == pd.Series):
            data = X.values
        else:
            data = X
        if len(data.shape) == 1:
            data = np.array([data])
        assert data.shape[1] == len(self.feature_names)

        variables = {}
        for x in self.feature_names:
            i = self.feature_indices[x]
            variables[x] = data[:, i]
        y_pred = self.evolved_phenotype.evaluate(**variables)

        return y_pred



from __future__ import annotations

from geneticengine.core.grammar import Grammar
from geneticengine.core.random.sources import Source
from geneticengine.core.representations.tree.utils import GengyList
from geneticengine.core.utils import build_finalizers
from geneticengine.metahandlers.base import MetaHandlerGenerator


class ListSizeBetween(MetaHandlerGenerator):
    """
    ListSizeBetween(a,b) restricts lists to be of length between a and b and implements a special list mutation.
    The list of options can be dynamically altered before the grammar extraction (Set.__annotations__["set"] = Annotated[List[Type], ListSizeBetween(c,d)].
    The special list mutation entails three different alterations to the list in question: deletion of a random element; addition of a random element; and replacement of a random element.
    """

    def __init__(self, min, max):
        self.min = min
        self.max = max

    def generate(
        self,
        r: Source,
        g: Grammar,
        rec,
        new_symbol,
        depth: int,
        base_type,
        ctx: dict[str, str],
    ):
        base_type = base_type.__args__[0]
        size = r.randint(self.min, self.max, str(base_type))
        fins = build_finalizers(lambda *x: rec(GengyList(base_type, x)), size)
        ident = ctx["_"]
        for i, fin in enumerate(fins):
            nctx = ctx.copy()
            nident = ident + "_" + str(i)
            nctx["_"] = nident
            new_symbol(base_type, fin, depth, nident, nctx)

    def mutate(
        self,
        r: Source,
        g: Grammar,
        random_node,
        depth: int,
        base_type,
        method,
        current_node,
    ):
        mutation_method = r.randint(0, 1)
        if (mutation_method == 0) and (len(current_node) != self.min):  # del
            element_to_be_deleted = r.randint(0, len(current_node) - 1)
            current_node.remove(current_node[element_to_be_deleted])
            return current_node
        elif (mutation_method == 1) and (len(current_node) != self.max):  # add
            new_element = random_node(r, g, depth, base_type.__args__[0], method=method)
            current_node.append(new_element)
            return current_node
        else:  # replace
            element_to_be_replaced = r.randint(0, len(current_node) - 1)
            new_element = random_node(r, g, depth, base_type.__args__[0], method=method)
            current_node[element_to_be_replaced] = new_element
            return current_node

    def crossover(
        self,
        r: Source,
        g: Grammar,
        options,
        arg,
        list_type,
        current_node,
    ):
        if not options or (len(current_node) < 2):
            return current_node
        n_elements_replaced = r.randint(1, len(current_node) - 1)
        big_enough_options = [
            getattr(o, arg)
            for o in options
            if len(getattr(o, arg)) >= n_elements_replaced
        ]
        while not big_enough_options:
            if n_elements_replaced == 1:
                return GengyList(list_type, current_node)
            n_elements_replaced = r.randint(1, n_elements_replaced - 1)
            big_enough_options = [
                getattr(o, arg)
                for o in options
                if len(getattr(o, arg)) >= n_elements_replaced
            ]
        option = big_enough_options[r.randint(0, len(big_enough_options) - 1)]

        # Always cut beginning as we do double crossovers, first using one tree as the current node, and then the second tree as current node.
        new_node = option[0:n_elements_replaced] + current_node[n_elements_replaced:]
        return GengyList(list_type, new_node)

    def __class_getitem__(self, args):
        return ListSizeBetween(*args)

    def __repr__(self):
        return f"ListSizeBetween[{self.min}...{self.max}]"


class ListSizeBetweenWithoutListOperations(MetaHandlerGenerator):
    """
    ListSizeBetweenWithoutListOperations(a,b) restricts lists to be of length between a and b.
    The list of options can be dynamically altered before the grammar extraction (Set.__annotations__["set"] = Annotated[List[Type], ListSizeBetweenWithoutListOperations(c,d)].
    """

    def __init__(self, min, max):
        self.min = min
        self.max = max

    def generate(
        self,
        r: Source,
        g: Grammar,
        rec,
        new_symbol,
        depth: int,
        base_type,
        ctx: dict[str, str],
    ):
        base_type = base_type.__args__[0]
        size = r.randint(self.min, self.max, str(base_type))
        fins = build_finalizers(lambda *x: rec(GengyList(base_type, x)), size)
        ident = ctx["_"]
        for i, fin in enumerate(fins):
            nctx = ctx.copy()
            nident = ident + "_" + str(i)
            nctx["_"] = nident
            new_symbol(base_type, fin, depth, nident, nctx)

    def __class_getitem__(self, args):
        return ListSizeBetweenWithoutListOperations(*args)

    def __repr__(self):
        return f"ListSizeBetweenWithoutListOperations[{self.min}...{self.max}]"



from __future__ import annotations


def fix_logs(s: str):
    s = s.replace("np.log(1 + np.abs", "log(")
    return s


def fix_sqrts(s: str):
    s = s.replace("np.sqrt(np.abs", "sqrt(")
    return s


def fix_numpy(s: str):
    s = s.replace("np.", "")
    return s


def fix_all(s: str):
    s = fix_numpy(fix_sqrts(fix_logs(s)))
    return s




import glob
import numpy as np
import pandas as pd

from geneticengine.exceptions import GeneticEngineError

def load(folder_name: str, x_axis: str, y_axis: str):
    print(f"Loading from: {folder_name}")
    f_list = glob.glob(f"{folder_name}/*.csv")

    data = list()

    for f in f_list:
        df = pd.read_csv(f)
        try:
            df = df[[x_axis, y_axis]]
            data.append(df)
        except:
            continue
    
    if not data:
        raise GeneticEngineError(f"No files in folder {folder_name} have both columns ({x_axis} and {y_axis}). \n We recommend using the CSVCallback in the GP class to generate the csv files with the GP class parameter save_to_csv.")
    all_data = pd.concat(data, axis=0, ignore_index=True)
    
    return all_data

def load_w_extra(folder_name: str, x_axis: str, y_axis:str, extra_column: str):
    print(f"Loading from: {folder_name}")
    f_list = glob.glob(f"{folder_name}/*.csv")

    data = list()

    for f in f_list:
        df = pd.read_csv(f)
        try:
            df = df[[x_axis, y_axis, extra_column]]
            data.append(df)
        except:
            continue
    
    if not data:
        raise GeneticEngineError(f"No files in folder {folder_name} have both columns ({x_axis}, {y_axis} and {extra_column}). \n We recommend using the CSVCallback in the GP class to generate the csv files with the GP class parameter save_to_csv.")
    all_data = pd.concat(data, axis=0, ignore_index=True)
    
    return all_data




from __future__ import annotations

from typing import Any
from typing import Callable
from typing import Dict
from typing import Type
from typing import TypeVar

from geneticengine.core.grammar import Grammar
from geneticengine.core.random.sources import Source
from geneticengine.metahandlers.base import MetaHandlerGenerator

min = TypeVar("min", covariant=True)
max = TypeVar("max", covariant=True)


class FloatRange(MetaHandlerGenerator):
    """
    FloatRange(a,b) restricts floats to be between a and b.
    The range can be dynamically altered before the grammar extraction (Float.__annotations__["value"] = Annotated[float, FloatRange(c,d)].
    """

    def __init__(self, min, max):
        self.min = min
        self.max = max

    def generate(
        self,
        r: Source,
        g: Grammar,
        rec,
        new_symbol,
        depth: int,
        base_type,
        context: dict[str, str],
    ):
        rec(r.random_float(self.min, self.max, str(base_type)))

    def __class_getitem__(self, args):
        return FloatRange(*args)

    def __repr__(self):
        return f"[{self.min},{self.max}]"


class FloatList(MetaHandlerGenerator):
    """
    FloatList([a_1, .., a_n]) restricts floats to be an element from the list [a_1, .., a_n].
    The range can be dynamically altered before the grammar extraction (Float.__init__.__annotations__["value"] = Annotated[float, FloatList[a_1, .., a_n]].
    """

    def __init__(self, elements):
        self.elements = elements

    def generate(
        self,
        r: Source,
        g: Grammar,
        rec,
        new_symbol,
        depth: int,
        base_type,
        context: dict[str, str],
    ):
        rec(r.choice(self.elements, str(base_type)))

    def __class_getitem__(self, args):
        return FloatList(*args)

    def __repr__(self):
        return f"[{self.elements}]"



from __future__ import annotations

from geneticengine.algorithms.gp.gp import GP  # noqa
from geneticengine.core.decorators import abstract  # noqa
from geneticengine.core.grammar import extract_grammar  # noqa
from geneticengine.core.random.sources import RandomSource  # noqa
from geneticengine.core.representations.grammatical_evolution.ge import (
    ge_representation,
)  # noqa
from geneticengine.core.representations.grammatical_evolution.structured_ge import (
    sge_representation,
)  # noqa
from geneticengine.core.representations.tree.treebased import random_node
from geneticengine.core.representations.tree.treebased import treebased_representation
from geneticengine.metahandlers.floats import FloatRange  # noqa
from geneticengine.metahandlers.ints import IntRange  # noqa
from geneticengine.metahandlers.lists import ListSizeBetween  # noqa
from geneticengine.metahandlers.vars import VarRange  # noqa
